/* $Id: iflow.c,v 1.10 2002/11/05 10:15:56 aa5779 Exp $ */
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#undef getc
#undef putc
#undef puts
#include "interp.h"
#include "flow.h"
#define IFLOW_IMP
#include "iflow.h"
#include "util.h"
#include "dstring.h"

flow_class inline_flow_class = {"inline", ilf_open, ilf_close, ilf_getc, ilf_ungetc, 
							   ilf_putc, ilf_puts,
							   ilf_getpos, ilf_setpos, ilf_rewind, ilf_flowctl
							 };

typedef struct inline_flow_data
{
	unsigned char writeable : 1;
	char *base;
	char *pos;
	variable *dest;
	unsigned char escape_zero : 1;
} inline_flow_data;

flow_handle_t ilf_open(char *name, char *opts, int writeable)
{
	inline_flow_data *ifd = ALLOC(inline_flow_data);

	ifd->writeable = writeable;
	if(writeable)
	{
		char *sep = strchr(name, '/');
		char *poolname;
		variable *dest;
		if(!sep)
			recover_error("iflow: can't parse shared variable name: %s", name);
		poolname = xmalloc(sep - name + 1);
		memcpy(poolname, name, sep - name);
		poolname[sep - name] = '\0';
		dest = get_shared_value(poolname, sep + 1);
		if(!dest)
			recover_error("iflow: shared variable '%s' undefined", name);
		free(poolname);
		ifd->dest = dest;
	}
	else
	{
		ifd->base = xstrdup(name);
		ifd->pos = ifd->base;
	}
	ifd->escape_zero = 0;
	return ifd;
}

void ilf_close(flow_handle_t fh)
{
	inline_flow_data *ifd = fh;
	if(!ifd->writeable)
		free(ifd->base);
	FREE(ifd);
}

int ilf_getc(flow_handle_t fh)
{
	inline_flow_data *ifd = fh;
	if(ifd->escape_zero && *ifd->pos == '\1')
	{
		ifd->pos++;
		if(*ifd->pos == '\1' || *ifd->pos == '\2')
			return (*ifd->pos++) - 1;
		recover_error("iflow: invalid escapez sequence: \\1\\%x", *ifd->pos);
	}
	return *ifd->pos ? *ifd->pos++ : FLOW_EOF;
}

int ilf_ungetc(flow_handle_t fh, int ch)
{
	inline_flow_data *ifd = fh;
	if(ch != FLOW_EOF)
	{
		if(ifd->escape_zero)
			recover_error("iflow: cannot ungetc when in escapez mode");
		if(ifd->pos == ifd->base)
			return FLOW_ERROR;
		*--ifd->pos = ch;
	}
	return 0;
}

int ilf_putc(flow_handle_t fh, int ch)
{
	inline_flow_data *ifd = fh;
	if(!ifd->writeable)
	{
		errno = EACCES;
		return FLOW_ERROR;
	}
	if(ifd->escape_zero && (!ch || ch == 1))
	{
		ds_appendch(ifd->dest->vals->val, '\1');
		ch++;
	}
	if(!ch)
		recover_error("iflow: attempt to store null char not in escapez mode");
	ds_appendch(ifd->dest->vals->val, ch);
	return 0;
}

int ilf_puts(flow_handle_t fh, char *str)
{
	inline_flow_data *ifd = fh;
	if(!ifd->writeable)
	{
		errno = EACCES;
		return FLOW_ERROR;
	}
	ds_appendstr(ifd->dest->vals->val, str);
	return 0;
}
 
flow_pos_t ilf_getpos(flow_handle_t fh)
{
	inline_flow_data *ifd = fh;
	return ifd->writeable ? ds_length(ifd->dest->vals->val) : ifd->pos - ifd->base;
}

int ilf_setpos(flow_handle_t fh, flow_pos_t fp)
{
	inline_flow_data *ifd = fh;
	if(ifd->writeable)
	{
		if(fp < 0 || fp > ds_length(ifd->dest->vals->val))
		{
			errno = EINVAL;
			return FLOW_ERROR;
		}
		ds_setsubstr(ifd->dest->vals->val, fp, ds_length(ifd->dest->vals->val) - fp, ds_create(NULL));
	}
	else
	{
		if(fp < 0 || fp > strlen(ifd->base))
		{
			errno = EINVAL;
			return FLOW_ERROR;
		}
		ifd->pos = ifd->base + fp;
	}
	return 0;
}

int ilf_rewind(flow_handle_t fh)
{
	inline_flow_data *ifd = fh;
	if(ifd->writeable)
		ds_assign(ifd->dest->vals->val, NULL);
	else
		ifd->pos = ifd->base;
	return 0;
}

char *ilf_flowctl(flow_handle_t fh, char *ctl)
{
	inline_flow_data *ifd = fh;
	if(!strcmp(ctl, "escapez"))
	{
		ifd->escape_zero = 1;
		return "";
	}
	else if(!strcmp(ctl, "noescapez"))
	{
		ifd->escape_zero = 0;
		return "";
	}
	else if(!strcmp(ctl, "escapez?"))
		return ifd->escape_zero ? "escapez" : "noescapez";
	else if(!strcmp(ctl, "loaded?"))
		return "true";
	else if(!strcmp(ctl, "flush") || !strcmp(ctl, "unbuffer") ||
			!strcmp(ctl, "lbuffer") || !strcmp(ctl, "buffer") ||
			!strcmp(ctl, "lock") || !strcmp(ctl, "unlock") || !strcmp(ctl, "lock?"))
		return "";
	return NULL;
}
